#include "kll0.h"
#include "chr.cpp"

extern "C" {
	#include "bla.h"
	int bla_main();
};

enum { EAX = 0, ECX = 1, EDX = 2, EBX = 3, EBP = 5, EDI = 7 };

struct ProgLine {
	int ln, qc;
	KSizPtr src;
};

int cmpProgLine(ProgLine *a, ProgLine *b) { return a->ln - b->ln; }

struct Prog {
	KIndexS ln;
	KSizPtr name;
	Prog() : ln(kmemPool, (void *) cmpProgLine), name(kmemPool) {}
	static void *dest(Prog *p) { KIndexS::dest(&p->ln); KSizPtr::dest(&p->name); return p; }
};

struct Env {
	KLexer *lx;
	KSizPtr *varSp;
	unsigned char *code;
	int qc, err, ln, timemode;
	KSizPtr stkFor, stkIf, gotoLink, prog;
	// fix邽߂spKv.
};

Env *env;

struct EnvLocal {
	int progselm;	// ̔ԍ.
	EnvLocal *envl0;
	int qc0;
	KSizPtr *pUsrStk;
};

EnvLocal *envl;

KIndexS *prog_getPln(int psel)
{
	if (env->prog.s <= psel * (int) sizeof (Prog *)) return 0;
	if (env->prog.getPtr(psel) == 0) return 0;
	return &((Prog *) env->prog.getPtr(psel))->ln;
}

//KIndexS *pln;
bla_Window *gwin;
static int col8[8] = { 0x000000, 0x0000ff, 0x00ff00, 0x00ffff, 0xff0000, 0xff00ff, 0xffff00, 0xffffff };
unsigned char gchr[128*128], gcol[128*128];
int chrofs_x = 0, chrofs_y = 0, progseln = 0;

void sub_print(int i)					{ printf("%d", i); }
void sub_printStr(int l, const char *s)	{ printf("%.*s", l, s); }
void sub_exit()							{ exit(0); }
void sub_input(int *p)					{ scanf("%d", p); }
void sub_rand(int *p)					{ *p = rand(); }
void sub_gcls(int c)					{ int i, *p = gwin->buf, xys = gwin->xsiz * gwin->ysiz; for (i = 0; i < 128 * 128; i++) gchr[i] = 0; for (i = 0; i < 128 * 128; i++) gcol[i] = 0; c = col8[c & 7]; for (i = 0; i < xys; i++) p[i] = c; }
void sub_openwin(int sx, int sy)		{ bla_closeWin(gwin); gwin = bla_openWin(sx, sy, "graphics"); sub_gcls(0); }
void sub_setpix(int x, int y, int c)	{ bla_setPixEx(gwin, x, y, c); }
void sub_flushwin()						{ bla_flushAll(gwin); }
void sub_leapflushwin(int msec)			{ bla_leapFlushAll(gwin, msec); }
void sub_ewait(int msec)				{ bla_flushAll(gwin); bla_waitEx(msec); }
void sub_einkey(int *p)					{ *p = bla_inkey(gwin) & ~ bla_KEY_LOCKS; }
void sub_getchr(int x, int y, int *p)	{ *p = gchr[y * 128 + x]; }
void sub_getcol(int x, int y, int *p)	{ *p = gcol[y * 128 + x]; }
void sub_chrofs(int x, int y)			{ chrofs_x = x; chrofs_y = y; }
void sub_gprinti(int x, int y, int w, int i, int c, int b) { char t[200]; sprintf(t, "%*d", w, i); bla_drawStrEx(gwin, x, y, col8[c], col8[b], t); }
void sub_gprints(int x, int y, int l, const char *s, int c, int b) { char t[200]; memcpy(t, s, l); t[l] = 0; bla_drawStrEx(gwin, x, y, col8[c], col8[b], t); }
void sub_progseln(int *p)				{ *p = progseln; }
void sub_strnew(KSizPtr **p)			{ *p = NEW0(kmemPool, KSizPtr)(kmemPool); }
void sub_strdel(KSizPtr *p)				{ KSizPtr::dest(p); kmemPool->free(sizeof (KSizPtr), p); }
void sub_printstr(KSizPtr *p)			{ printf("%.*s", p->s, p->p); }
void sub_strcpyimm(KSizPtr *p, int l, char *s) { p->reserve(l); memcpy(p->p, s, l); p->s = l; }
void sub_arysizs(KSizPtr *a, int s)		{ a->reserve0(s * sizeof (int)); a->s = s * sizeof (int); }
void sub_aryset(KSizPtr *a, unsigned int i, int v)  { if (i < a->s * sizeof (int)) a->setInt(i, v); }
void sub_aryget(KSizPtr *a, unsigned int i, int *v) { if (i < a->s * sizeof (int)) *v = a->getInt(i); }
void sub_timemode(int i)				{ env->timemode = i; }

void sub_mem()
{
	int n, u = kmemPool->inUse(&n);
	printf("memory in use: %d (n=%d)\n", u, n);
}

void sub_list(int l0, int l1)
{
	l0 = 0;
	l1 = 0x7fffffff;
	KIndexS *pn = prog_getPln(progseln);
	if (pn == 0) return;
	int cr, i = pn->search(&l0, &cr);
	while (i < pn->sp.s) {
		ProgLine *pl = *(ProgLine **) &pn->sp.p[i];
		if (pl->ln > l1) break;
		printf("%4d %.*s\n", pl->ln, pl->src.s, pl->src.p);
		i += sizeof (void *);
	}
}

int sub_initstack(int s)
{
	if (s > 0) {
		envl->pUsrStk->reserve(s * 4);
		envl->pUsrStk->s = 0;
	}
	return (int) envl->pUsrStk->p;
}

void sub_echr(int x, int y, int ch, int c, int b)
{
	x += chrofs_x;
	y += chrofs_y;
	if (0 <= x && x * 16 < gwin->xsiz && 0 <= y && y * 16 < gwin->ysiz && ch < 6) {
		gchr[y * 128 + x] = ch;
		gcol[y * 128 + x] = c;
		int i, j;
		char *p = &chr[ch * 256];
		x *= 16;
		y *= 16;
		int *q = gwin->buf + x + y * gwin->xsiz;
		c = col8[c & 7];
		b = col8[b & 7];
		for (j = 0; j < 16; j++) {
			for (i = 0; i < 16; i++) {
				q[i] = b;
				if (p[i] == '*')
					q[i] = c;
			}
			p += 16;
			q += gwin->xsiz;
		}
	}
}

void sub_chrbox(int sx, int sy, int x, int y, int ch, int c, int b)
{
	for (int j = 0; j < sy; j++) {
		for (int i = 0; i < sx; i++)
			sub_echr(x + i, y + j, ch, c, b);
	}
}

void sub_scroll(int d)
{
	int x, y, *b = gwin->buf, xs = gwin->xsiz, ys;
	if (d == 0) {	/*  */
		for (y = 0; y < gwin->ysiz; y++) {
			for (x = 0; x < xs - 16; x++)
				b[x] = b[x + 16];
			b += xs;
		}
		xs = (xs >> 4) - 1;
		ys = gwin->ysiz >> 4;
		for (y = 0; y < ys; y++) {
			for (x = 0; x < xs; x++) {
				gchr[y * 128 + x] = gchr[y * 128 + x + 1];
				gcol[y * 128 + x] = gcol[y * 128 + x + 1];
			}
		}
	}
	if (d == 1) {	/*  */
		for (y = 0; y < gwin->ysiz; y++) {
			for (x = xs - 1; x >= 16; x--)
				b[x] = b[x - 16];
			b += xs;
		}
		xs >>= 4;
		ys = gwin->ysiz >> 4;
		for (y = 0; y < ys; y++) {
			for (x = xs - 1; x >= 1; x--) {
				gchr[y * 128 + x] = gchr[y * 128 + x - 1];
				gcol[y * 128 + x] = gcol[y * 128 + x - 1];
			}
		}
	}
	if (d == 2) {	/*  */
		int xys = gwin->xsiz * (gwin->ysiz - 16);
		xs <<= 4;
		for (y = 0; y < xys; y++)
			b[y] = b[y + xs];
		xys = ((gwin->ysiz >> 4) - 1) * 128;
		for (y = 0; y < xys; y++) {
			gchr[y] = gchr[y + 128];
			gcol[y] = gcol[y + 128];
		}
	}
	if (d == 3) {	/*  */
		int xys = gwin->xsiz * (gwin->ysiz - 16);
		xs <<= 4;
		for (y = xys - 1; y >= xs; y--)
			b[y] = b[y - xs];
		xys = ((gwin->ysiz >> 4) - 1) * 128;
		for (y = xys - 1; y >= 128; y--) {
			gchr[y] = gchr[y - 128];
			gcol[y] = gcol[y - 128];
		}
	}
}

void sub_progsel(int i)
{
	while (env->prog.s <= i * (int) sizeof (Prog *))
		env->prog.addPtr(0);
	if (env->prog.getPtr(i) == 0)
		*((Prog **) &env->prog.p[i * (int) sizeof (Prog *)]) = NEW0(kmemPool, Prog);
//	pln = &((Prog *) env->prog.getPtr(i))->ln;
	progseln = i;
}

void sub_proglist()
{
	int i, i1 = env->prog.s / sizeof (Prog *);
	for (i = 0; i < i1; i++) {
		Prog *pr = (Prog *) env->prog.getPtr(i);
		if (pr != 0)
			printf("  #%02d [%d]\n", i, pr->ln.sp.s / sizeof (void *));
	}
	printf ("  progseln=%02d\n", progseln);
}

void sub_new(int psel)
{
	KIndexS *pn = prog_getPln(psel);
	if (pn == 0) return;
	while (pn->sp.s > 0) {
		int i = pn->sp.s - sizeof (void *);
		ProgLine *pl = *(ProgLine **) &pn->sp.p[i];
		KSizPtr::dest(&pl->src);
		kmemPool->free(sizeof (ProgLine), pl);
		pn->sp.delPtr(i);
	}
}

void sub_progdel(int i)
{
	if (i == progseln) return;
	if (prog_getPln(i) == 0) return;
	sub_new(i);
	Prog::dest((Prog *) env->prog.getPtr(i));
	kmemPool->free(sizeof (Prog), env->prog.getPtr(i));
	*((Prog **) &env->prog.p[i * (int) sizeof (Prog *)]) = 0;	// setPtr.
}

void sub_progget(int *plin, KSizPtr *sp, int lln, int psel)
{
	*plin = -1;
	sp->s = 0;
	KIndexS *ln = prog_getPln(psel);
	if (ln == 0) return;
	if (ln->sp.s <= lln * (int) sizeof (ProgLine *)) return;
	ProgLine *pl = (ProgLine *) ln->sp.getPtr(lln);
	*plin = pl->ln;
	sp->reserve(pl->src.s);
	memcpy(sp->p, pl->src.p, pl->src.s);
	sp->s = pl->src.s;
}

void sub_proglin1(int *plln1, int psel)
{
	*plln1 = -1;
	KIndexS *ln = prog_getPln(psel);
	if (ln == 0) return;
	*plln1 = ln->sp.sizPtr();
}

void prog_delLine(KIndexS *pn, int lin)
{
	int i, cr;
	i = pn->search(&lin, &cr);
	if (cr == 0) {
		ProgLine *pl = *(ProgLine **) &pn->sp.p[i];
		KSizPtr::dest(&pl->src);
		kmemPool->free(sizeof (ProgLine), pl);
		pn->sp.delPtr(i);
	}
}

void prog_insLine(KIndexS *pn, int lin, int s, char *p)
{
	ProgLine *pl = new(kmemPool->alloc(sizeof (ProgLine))) ProgLine;
	pl->ln = lin;
	pl->src.addBytes(s, p);
	pn->ins(pl);
}

void sub_progputimm(int ln, int l, char *s, int psel)
{
	KIndexS *pn = prog_getPln(psel);
	if (pn != 0) {
		prog_delLine(pn, ln);
		if (l > 0)
			prog_insLine(pn, ln, l, s);
	}
}

void sub_progput(int ln, KSizPtr *sp, int psel)
{
	sub_progputimm(ln, sp->s, sp->p, psel);
}

void sub_prognew(int *p)
{
	int i, i1 = env->prog.s / (int) sizeof (Prog *);
	for (i = 0; i < i1; i++) {
		if (env->prog.getPtr(i) == 0) break;
	}
	if (i >= i1)
		env->prog.addPtr(0);
	*((Prog **) &env->prog.p[i * (int) sizeof (Prog *)]) = NEW0(kmemPool, Prog);	// setPtr
	*p = i;
}

int sub_run(int psel);

struct StkFor {
	int qc, c, c1, d;
};

struct StkIf {
	int qc0, qc1;
};

int opCallEx(unsigned char *q, KLexer_Elmt *le, const char *s,...)
{
	int qc = 0, i = 0;
    va_list ap;
    va_start(ap, s);
	void *p = va_arg(ap, void *);
	while (*s != '\0') {
		char c = *s++;
		if (c == 'c')
			qc += kputBin(&q[qc], "b8_%i32", va_arg(ap, int));
		if (c == 'v')
			qc += kputBin(&q[qc], "8b_85_%i32", va_arg(ap, int) * 4);
		if (c == 'V')
			qc += kputBin(&q[qc], "8b_85_%i32", le[va_arg(ap, int)].i * 4);
		if (c == 'p')
			qc += kputBin(&q[qc], "8d_85_%i32", va_arg(ap, int) * 4);
		if (c == 'P')
			qc += kputBin(&q[qc], "8d_85_%i32", le[va_arg(ap, int)].i * 4);
		if (i == 0)
			qc += kputBin(&q[qc], "89_04_24");
		else
			qc += kputBin(&q[qc], "89_44_24_%i8", i * 4);
		i++;
	}
    va_end(ap);
	qc += kputBin(&q[qc], "e8_%r32", p);
	return qc;
}

void eval0(int psel)
{
	env->err = 0;
	EnvLocal *envl0 = envl;
	envl = NEW0(kmemPool, EnvLocal);
	envl->envl0 = envl0;
	envl->progselm = psel;
	envl->qc0 = env->qc;
	envl->pUsrStk = NEW0(kmemPool, KSizPtr)(kmemPool);
	env->qc += kputBin(&env->code[env->qc], "60  83_ec_7c"); // PUSHAD.  SUB ESP,124.  ̖߂ŃX^bN𒲐.
	env->qc += kputBin(&env->code[env->qc], "%i8_%i32", 0xb8 + EBP, 0);
	// [Jϐ̂Ƃ͏l. ͍lȂ.
	env->qc += opCallEx(&env->code[env->qc], 0, "c", sub_initstack, 1024);	// ftHg.
	env->qc += kputBin(&env->code[env->qc], "97");
	env->stkFor.s = 0;
	env->stkIf.s = 0;
	env->gotoLink.s = 0;
}

int evalAdd(int l, const char *s, KIndexS *pn)
{
	static int ic[99][32];
	if (env->err != 0) return 1;
	if (s[l - 1] == '\n')
		l--;
	env->lx->sp->s = 0;
	KStringId *sid = env->lx->sid;
	int i0 = *sid->n, pc = 0;
	env->lx->add(l, (const char *) s);
	KLexer_Elmt *le = (KLexer_Elmt *) env->lx->add(1, ";")->p; 
	int	i1 = *sid->n, *var = (int *) env->varSp->reserve(i1 * sizeof (int))->p;
	for (int i = i0; i < i1; i++)
		var[i] = env->lx->getConstInt(i);	// intł͂ȂƂ0Ԃ.
   	for (int n = env->lx->getElmtLen(); pc < n; pc++) {
		if (kphrCmp0(32, ic[ 0], "@ ;", &le[pc], sid)) continue;
		if (kphrCmp0(32, ic[ 1], "@ @* =", &le[pc], sid)) { // 2Pڂ"=".
			if (kphrCmp0(32, ic[ 2], "@ @* = @* ;", &le[pc], sid)) {	// P.
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  89_85_%i32", EAX * 8 + 0x85, le[pc + 2].i * 4, le[pc].i * 4);
			} else if (kphrCmp0(32, ic[ 3], "@ @* = @* + @* ;", &le[pc], sid)) { // Z.
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  01_c8  89_85_%i32", EAX * 8 + 0x85, le[pc + 2].i * 4, ECX * 8 + 0x85, le[pc + 4].i * 4, le[pc].i * 4);
			} else if (kphrCmp0(32, ic[ 4], "@ @* = @* - @* ;", &le[pc], sid)) { // Z.
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  29_c8  89_85_%i32", EAX * 8 + 0x85, le[pc + 2].i * 4, ECX * 8 + 0x85, le[pc + 4].i * 4, le[pc].i * 4);
			} else if (kphrCmp0(32, ic[ 5], "@ @* = @* % @* ;", &le[pc], sid)) { // ]Z.
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  99  f7_f9  89_%i8_%i32", EAX * 8 + 0x85, le[pc + 2].i * 4, ECX * 8 + 0x85, le[pc + 4].i * 4, EDX * 8 + 0x85, le[pc].i * 4);
			} else if (kphrCmp0(32, ic[ 6], "@ @* = @* * @* ;", &le[pc], sid)) { // Z.
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  0f_af_c1   89_%i8_%i32", EAX * 8 + 0x85, le[pc + 2].i * 4, ECX * 8 + 0x85, le[pc + 4].i * 4, EAX * 8 + 0x85, le[pc].i * 4);
			} else if (kphrCmp0(32, ic[ 7], "@ @* = @* / @* ;", &le[pc], sid)) { // Z.
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  99  f7_f9  89_%i8_%i32", EAX * 8 + 0x85, le[pc + 2].i * 4, ECX * 8 + 0x85, le[pc + 4].i * 4, EAX * 8 + 0x85, le[pc].i * 4);
			} else if (kphrCmp0(32, ic[ 8], "@ @* = @* *>> @* , @* ;", &le[pc], sid)) { // Œ菬␳Z.
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  f7_e9  0f_ac_d0_%i8  89_85_%i32",
					EAX * 8 + 0x85, le[pc + 2].i * 4, ECX * 8 + 0x85, le[pc + 4].i * 4, env->lx->getConstInt(le[pc + 6].i), le[pc].i * 4);
			} else
				goto err;
		} else if (kphrCmp0(32, ic[ 9], "@ PRINT", &le[pc], sid)) { // print.
			int lpc = pc + 1;
			for (;;) {
				if (kphrCmp0(32, ic[10], "@ ;", &le[lpc], sid)) {
					env->qc += opCallEx(&env->code[env->qc], &le[pc], "cc", sub_printStr, 1, "\n");
					break;
				}
				if (kphrCmp0(32, ic[11], "@ , ;", &le[lpc], sid)) {
					env->qc += opCallEx(&env->code[env->qc], &le[pc], "cc", sub_printStr, 1, " ");
					break;
				}
				if (kphrCmp0(32, ic[12], "@ ,", &le[lpc], sid)) {
					env->qc += opCallEx(&env->code[env->qc], &le[pc], "cc", sub_printStr, 1, " ");
					lpc++;
					continue;
				}
				if (*le[lpc].p0 == 0x22 || *le[lpc].p0 == 0x27) {
					env->qc += opCallEx(&env->code[env->qc], &le[pc], "cc", sub_printStr, sid->getStr(le[lpc].i)->l - 2, le[lpc].p0 + 1);
				} else {
					env->qc += opCallEx(&env->code[env->qc], &le[lpc], "V", sub_print, 0);
				}
				lpc++;
				if (kphrCmp0(32, ic[13], "@ , ;", &le[lpc], sid)) break;
				if (kphrCmp0(32, ic[14], "@ ,", &le[lpc], sid))
					lpc++;
			}
		} else if (kphrCmp0(32, ic[15], "@ EXIT ;", &le[pc], sid)) { // exit.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "", sub_exit);
		} else if (kphrCmp0(32, ic[16], "@ LIST ;", &le[pc], sid)) { // list.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "", sub_list);
		} else if (kphrCmp0(32, ic[17], "@ FOR @* = @* , @* ;", &le[pc], sid)) {	// for.
			StkFor st;
			env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  89_85_%i32", EAX * 8 + 0x85, le[pc + 3].i * 4, le[pc + 1].i * 4);
			st.qc = env->qc;
			st.c = le[pc + 1].i;
			st.c1 = le[pc + 5].i;
			st.d = -1;
			env->stkFor.addBytes(sizeof (StkFor), (char *) &st);
		} else if (kphrCmp0(32, ic[18], "@ FOR @* = @* , @* , @* ;", &le[pc], sid)) {	// for.
			StkFor st;
			env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  89_85_%i32", EAX * 8 + 0x85, le[pc + 3].i * 4, le[pc + 1].i * 4);
			st.qc = env->qc;
			st.c = le[pc + 1].i;
			st.c1 = le[pc + 5].i;
			st.d = le[pc + 7].i;
			env->stkFor.addBytes(sizeof (StkFor), (char *) &st);
		} else if (kphrCmp0(32, ic[19], "@ NEXT ;", &le[pc], sid)) {	// next
			if (env->stkFor.s < (int) sizeof (StkFor))
				goto err;
			StkFor *st = (StkFor *) &env->stkFor.p[env->stkFor.s - sizeof (StkFor)];
			env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32", EAX * 8 + 0x85, st->c * 4, ECX * 8 + 0x85, st->c1 * 4);
			if (st->d < 0)
				env->qc += kputBin(&env->code[env->qc], "40");
			else
				env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  03_c2", EDX * 8 + 0x85, st->d * 4);
			env->qc += kputBin(&env->code[env->qc], "89_85_%i32  39_c8  0f_8e_%r32", st->c * 4, &env->code[st->qc]);
			env->stkFor.s -= sizeof (StkFor);
		} else if (kphrCmp0(32, ic[20], "@ MEM ;", &le[pc], sid)) { // mem.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "", sub_mem);
		} else if (kphrCmp0(32, ic[21], "@ INPUT @* ;", &le[pc], sid)) { // input.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_input, 1);
		} else if (kphrCmp0(32, ic[50], "@ INPUT @* , @* ;", &le[pc], sid) != 0 && (*le[pc + 1].p0 == 0x22 || *le[pc + 1].p0 == 0x27)) { // input.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "cc", sub_printStr, sid->getStr(le[pc + 1].i)->l - 2, le[pc + 1].p0 + 1);
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_input, 3);
		} else if (kphrCmp0(32, ic[22], "@ IF @* @* @* THEN", &le[pc], sid) != 0 && 16 <= le[pc + 2].i && le[pc + 2].i <= 21) {	// if ... then.
			static const unsigned char t[6] = { 0x84, 0x85, 0x8c, 0x8f, 0x8e, 0x8d }; // ==(JE), !=(JNE), <(JL), >(JG), <=(JLE), >=(JGE).
			StkIf st;
			env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  39_c8", EAX * 8 + 0x85, le[pc + 1].i * 4, ECX * 8 + 0x85, le[pc + 3].i * 4);
			env->qc += kputBin(&env->code[env->qc], "0f_%i8_%r32", t[le[pc + 2].i - 16] ^ 1, 0);
			st.qc0 = env->qc - 4;
			st.qc1 = -1;
			env->stkIf.addBytes(sizeof (StkIf), (char *) &st);
			pc += 5 - 1;
			continue;
		} else if (kphrCmp0(32, ic[23], "@ ELSE", &le[pc], sid)) {	// else.
			if (env->stkIf.s < (int) sizeof (StkIf))
				goto err;
			StkIf *st = (StkIf *) &env->stkIf.p[env->stkIf.s - sizeof (StkIf)];
			env->qc += kputBin(&env->code[env->qc], "e9_%r32", 0);
			st->qc1 = env->qc - 4;
			kputBin(&env->code[st->qc0], "%r32", &env->code[env->qc]);
			continue;
		} else if (kphrCmp0(32, ic[24], "@ FI ;", &le[pc], sid)) {	// fi.
			if (env->stkIf.s < (int) sizeof (StkIf))
				goto err;
			StkIf *st = (StkIf *) &env->stkIf.p[env->stkIf.s - sizeof (StkIf)];
			if (st->qc1 < 0)
				kputBin(&env->code[st->qc0], "%r32", &env->code[env->qc]);
			else
				kputBin(&env->code[st->qc1], "%r32", &env->code[env->qc]);
			env->stkIf.s -= sizeof (StkIf);
		} else if (kphrCmp0(32, ic[25], "@ GOTO @* ;", &le[pc], sid)) {	// goto.
			int i = env->lx->getConstInt(le[pc + 1].i), cr;
			i = pn->search(&i, &cr);
			if (cr != 0) goto err;
			env->qc += kputBin(&env->code[env->qc], "e9_%r32", 0);
			env->gotoLink.addInt(env->qc - 4)->addInt(*(int *) &pn->sp.p[i]);
		} else if (kphrCmp0(32, ic[26], "@ IF @* @* @* GOTO @* ;", &le[pc], sid) != 0 && 16 <= le[pc + 2].i && le[pc + 2].i <= 21) {	// if ... goto.
			static const unsigned char t[6] = { 0x84, 0x85, 0x8c, 0x8f, 0x8e, 0x8d }; // ==(JE), !=(JNE), <(JL), >(JG), <=(JLE), >=(JGE).
			env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32  8b_%i8_%i32  39_c8", EAX * 8 + 0x85, le[pc + 1].i * 4, ECX * 8 + 0x85, le[pc + 3].i * 4);
			env->qc += kputBin(&env->code[env->qc], "0f_%i8_%r32", t[le[pc + 2].i - 16], 0);
			int i = env->lx->getConstInt(le[pc + 5].i), cr;
			i = pn->search(&i, &cr);
			if (cr != 0) goto err;
			env->gotoLink.addInt(env->qc - 4)->addInt(*(int *) &pn->sp.p[i]);
		} else if (kphrCmp0(32, ic[27], "@ RAND @* ;", &le[pc], sid)) { // rand.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_rand, 1);
		} else if (kphrCmp0(32, ic[28], "@ OPENWIN @* , @* ;", &le[pc], sid)) {	// openwin.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VV", sub_openwin, 1, 3);
		} else if (kphrCmp0(32, ic[29], "@ FLUSHWIN ;", &le[pc], sid)) {	// flushwin.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "", sub_flushwin);
		} else if (kphrCmp0(32, ic[30], "@ SETPIX @* , @* , @* ;", &le[pc], sid)) {	// setpix.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVV", sub_setpix, 1, 3, 5);
		} else if (kphrCmp0(32, ic[31], "@ INITSTACK @* ;", &le[pc], sid)) {	// initstack.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_initstack, 1);
			env->qc += kputBin(&env->code[env->qc], "97");
		} else if (kphrCmp0(32, ic[32], "@ PUSH @* ;", &le[pc], sid)) {	// push.
			env->qc += kputBin(&env->code[env->qc], "8b_%i8_%i32   89_07  83_c7_04", EAX * 8 + 0x85, le[pc + 1].i * 4);
		} else if (kphrCmp0(32, ic[33], "@ POP @* ;", &le[pc], sid)) {	// pop.
			env->qc += kputBin(&env->code[env->qc], "83_c7_fc   8b_07   89_%i8_%i32", EAX * 8 + 0x85, le[pc + 1].i * 4);
		} else if (kphrCmp0(32, ic[34], "@ GOSUB @* ;", &le[pc], sid)) {	// gosub.
			int i = env->lx->getConstInt(le[pc + 1].i), cr;
			i = pn->search(&i, &cr);
			if (cr != 0) goto err;
			env->qc += kputBin(&env->code[env->qc], "b8_%i32  89_07  83_c7_04  e9_%r32", &env->code[env->qc + 15], 0);
			env->gotoLink.addInt(env->qc - 4)->addInt(*(int *) &pn->sp.p[i]);
			env->qc += kputBin(&env->code[env->qc], "83_c7_fc");
		} else if (kphrCmp0(32, ic[35], "@ RETURN ;", &le[pc], sid)) {	// return.
			env->qc += kputBin(&env->code[env->qc], "ff_67_fc");
		} else if (kphrCmp0(32, ic[36], "@ LEAPFLUSHWIN @* ;", &le[pc], sid)) {	// leapflushwin.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_leapflushwin, 1);
		} else if (kphrCmp0(32, ic[37], "@ GCLS ;", &le[pc], sid)) {	// gcls;
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "c", sub_gcls, 0);
		} else if (kphrCmp0(32, ic[38], "@ GCLS @* ;", &le[pc], sid)) {	// gcls #;
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_gcls, 1);
		} else if (kphrCmp0(32, ic[39], "@ ECHR @* , @* , @* ;", &le[pc], sid)) {	// echr.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVcc", sub_echr, 1, 3, 5, 7, 0);
		} else if (kphrCmp0(32, ic[40], "@ ECHR @* , @* , @* , @* ;", &le[pc], sid)) {	// echr.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVc", sub_echr, 1, 3, 5, 7, 0);
		} else if (kphrCmp0(32, ic[41], "@ ECHR @* , @* , @* , @* , @* ;", &le[pc], sid)) {	// echr.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVV", sub_echr, 1, 3, 5, 7, 9);
		} else if (kphrCmp0(32, ic[42], "@ EWAIT @* ;", &le[pc], sid)) {	// ewait.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_ewait, 1);
		} else if (kphrCmp0(32, ic[43], "@ EINKEY @* ;", &le[pc], sid)) { // einkey.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_einkey, 1);
		} else if (kphrCmp0(32, ic[44], "@ GETCHR @* , @* , @* ;", &le[pc], sid)) { // getchr.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVP", sub_getchr, 1, 3, 5);
		} else if (kphrCmp0(32, ic[45], "@ GETCOL @* , @* , @* ;", &le[pc], sid)) { // getcol.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVP", sub_getcol, 1, 3, 5);
		} else if (kphrCmp0(32, ic[46], "@ CHRBOX @* , @* , @* , @* , @* ;", &le[pc], sid)) { // chrbox.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVVcc", sub_chrbox, 1, 3, 5, 7, 9, 7, 0);
		} else if (kphrCmp0(32, ic[47], "@ CHRBOX @* , @* , @* , @* , @* , @* ;", &le[pc], sid)) { // chrbox.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVVVc", sub_chrbox, 1, 3, 5, 7, 9, 11, 0);
		} else if (kphrCmp0(32, ic[48], "@ CHRBOX @* , @* , @* , @* , @* , @* , @* ;", &le[pc], sid)) { // chrbox.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVVVV", sub_chrbox, 1, 3, 5, 7, 9, 11, 13);
		} else if (kphrCmp0(32, ic[49], "@ CHROFS @* , @* ;", &le[pc], sid)) { // chrofs.
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VV", sub_chrofs, 1, 3);
		} else if (kphrCmp0(32, ic[51], "@ GPRINTS @* , @* , @* ;", &le[pc], sid) != 0 && *le[pc + 5].p0 == 0x22) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVcccc", sub_gprints, 1, 3, sid->getStr(le[pc + 5].i)->l - 2, le[pc + 5].p0 + 1, 7, 0);
		} else if (kphrCmp0(32, ic[52], "@ GPRINTS @* , @* , @* , @* ;", &le[pc], sid) != 0 && *le[pc + 5].p0 == 0x22) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVccVc", sub_gprints, 1, 3, sid->getStr(le[pc + 5].i)->l - 2, le[pc + 5].p0 + 1, 7, 0);
		} else if (kphrCmp0(32, ic[53], "@ GPRINTS @* , @* , @* , @* , @* ;", &le[pc], sid) != 0 && *le[pc + 5].p0 == 0x22) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVccVV", sub_gprints, 1, 3, sid->getStr(le[pc + 5].i)->l - 2, le[pc + 5].p0 + 1, 7, 9);
		} else if (kphrCmp0(32, ic[54], "@ GPRINTI @* , @* , @* , @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVcc", sub_gprinti, 1, 3, 5, 7, 7, 0);
		} else if (kphrCmp0(32, ic[55], "@ GPRINTI @* , @* , @* , @* , @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVVc", sub_gprinti, 1, 3, 5, 7, 9, 0);
		} else if (kphrCmp0(32, ic[56], "@ GPRINTI @* , @* , @* , @* , @* , @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVVVVV", sub_gprinti, 1, 3, 5, 7, 9, 11);
		} else if (kphrCmp0(32, ic[57], "@ SCROLL @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_scroll, 1);
		} else if (kphrCmp0(32, ic[58], "@ PROGLIST ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "", sub_proglist);
		} else if (kphrCmp0(32, ic[59], "@ PROGSEL @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_progsel, 1);
		} else if (kphrCmp0(32, ic[60], "@ NEW ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "c", sub_new, progseln);
		} else if (kphrCmp0(32, ic[75], "@ NEW @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_new, 1);
		} else if (kphrCmp0(32, ic[61], "@ PROGSELN @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_progseln, 1);
		} else if (kphrCmp0(32, ic[62], "@ PROGDEL @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_progdel, 1);
		} else if (kphrCmp0(32, ic[63], "@ STRNEW @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_strnew, 1);
		} else if (kphrCmp0(32, ic[64], "@ STRDEL @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_strdel, 1);
		} else if (kphrCmp0(32, ic[65], "@ PRINTSTR @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_printstr, 1);
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "cc", sub_printStr, 1, "\n");
		} else if (kphrCmp0(32, ic[78], "@ PRINTSTR @* , ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_printstr, 1);
		} else if (kphrCmp0(32, ic[66], "@ STRCPYIMM @* , @* ;", &le[pc], sid) != 0 && *le[pc + 3].p0 == 0x22) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "Vcc", sub_strcpyimm, 1, sid->getStr(le[pc + 3].i)->l - 2, le[pc + 3].p0 + 1);
		} else if (kphrCmp0(32, ic[67], "@ PROGGET @* , @* , @* , @* ;", &le[pc], sid) != 0) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "PVVV", sub_progget, 1, 3, 5, 7);
		} else if (kphrCmp0(32, ic[68], "@ PROGGET @* , @* , @* ;", &le[pc], sid) != 0) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "PVVc", sub_progget, 1, 3, 5, progseln);
		} else if (kphrCmp0(32, ic[69], "@ PROGPUT @* , @* , @* ;", &le[pc], sid) != 0) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVV", sub_progput, 1, 3, 5);
		} else if (kphrCmp0(32, ic[70], "@ PROGPUT @* , @* ;", &le[pc], sid) != 0) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVc", sub_progput, 1, 3, progseln);
		} else if (kphrCmp0(32, ic[71], "@ PROGPUTIMM @* , @* , @* ;", &le[pc], sid) != 0 && (*le[pc + 3].p0 == 0x22 || *le[pc + 3].p0 == 0x27)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VccV", sub_progputimm, 1, sid->getStr(le[pc + 3].i)->l - 2, le[pc + 3].p0 + 1, 5);
		} else if (kphrCmp0(32, ic[72], "@ PROGPUTIMM @* , @*  ;", &le[pc], sid) != 0 && (*le[pc + 3].p0 == 0x22 || *le[pc + 3].p0 == 0x27)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "Vccc", sub_progputimm, 1, sid->getStr(le[pc + 3].i)->l - 2, le[pc + 3].p0 + 1, progseln);
		} else if (kphrCmp0(32, ic[73], "@ PROGLIN1 @* , @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "PV", sub_proglin1, 1, 3);
		} else if (kphrCmp0(32, ic[74], "@ PROGLIN1 @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "Pc", sub_proglin1, 1, progseln);
		} else if (kphrCmp0(32, ic[76], "@ PROGNEW @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_prognew, 1);
		} else if (kphrCmp0(32, ic[77], "@ PROGRUN @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_run, 1);
		} else if (kphrCmp0(32, ic[79], "@ ARYNEW @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "P", sub_strnew, 1);
		} else if (kphrCmp0(32, ic[80], "@ ARYDEL @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_strdel, 1);
		} else if (kphrCmp0(32, ic[81], "@ ARYSIZS @* , @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VV", sub_arysizs, 1, 3);
		} else if (kphrCmp0(32, ic[82], "@ ARYSET @* , @* , @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVV", sub_aryset, 1, 3, 5);
		} else if (kphrCmp0(32, ic[83], "@ ARYGET @* , @* , @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "VVP", sub_aryget, 1, 3, 5);
		} else if (kphrCmp0(32, ic[84], "@ TIMEMODE @* ;", &le[pc], sid)) {
			env->qc += opCallEx(&env->code[env->qc], &le[pc], "V", sub_timemode, 1);
		} else
			goto err;
		while (le[pc].i != 0)
			pc++;
	}
	return 0;
err:
	env->err = 1;
	if (l >= 30)
		l = 30;
	if (env->ln >= 0)
		printf("syntax error [%d] : %.*s\n\n", env->ln, l, le[pc].p0);
	else
		printf("syntax error : %.*s\n\n", l, le[pc].p0);
	return 1;
}

void eval1()
{
	if (env->stkFor.s != 0) {
		printf("nest error : FOR\n");
		goto fin;
	}
	if (env->stkIf.s != 0) {
		printf("nest error : IF\n");
		goto fin;
	}
	for (int i = 0; i < env->gotoLink.s; i += 2 * sizeof (int)) {
		int *p = (int *) &env->gotoLink.p[i];
		kputBin(&env->code[p[0]], "%r32", &env->code[((ProgLine *) p[1])->qc]);
	}
	env->qc += kputBin(&env->code[env->qc], "83_c4_7c  61  c3"); // ADD ESP,124.  ŏɒX^bNɖ߂.  POPAD.  RET.
	kputBin(&env->code[envl->qc0 + 5], "%i32", env->varSp->p);
	// for (int i = 0; i < qc; i++) { printf("%02x ", code[i]); }  //  fobOp.
	{
		void (*func)();	// func͊֐ւ̃|C^^̕ϐ.
		func = (void (*)()) (env->code + envl->qc0);
    	func();
	}
fin:
	env->qc = envl->qc0;
	KSizPtr::dest(envl->pUsrStk);
	kmemPool->free(sizeof (KSizPtr), envl->pUsrStk);
	EnvLocal *envl0 = envl->envl0;
	kmemPool->free(sizeof (EnvLocal), envl);
	envl = envl0;
}

int sub_run(int psel)
{
	KIndexS *pn = prog_getPln(psel);
	eval0(psel);
	int i, i1 = pn->sp.s / (int) sizeof (ProgLine *);
	for (i = 0; i < i1; i++) {
		ProgLine *pl = (ProgLine *) pn->sp.getPtr(i);
		env->ln = pl->ln; // G[\̂߂ɕKv.
		pl->qc = env->qc;
		if (evalAdd(pl->src.s, pl->src.p, pn) != 0) return 1;
	}
	eval1();
	return 0;
}

int bla_main(/* int argc, const char **argv */)
{
//	(void) argc;
//	(void) argv;
	KLexer lx(NEW1(kautoreleasePool, KStringId)(1, "; = + - PRINT EXIT LIST FOR , NEXT MEM INPUT IF THEN ELSE FI == != < > <= >= % * / GOTO RAND "
		"OPENWIN CLOSEWIN FLUSHWIN SETPIX INITSTACK PUSH POP GOSUB RETURN *>> LEAPFLUSHWIN GCLS ECHR EWAIT EINKEY GETCHR GETCOL "
		"CHRBOX CHROFS GPRINTS GPRINTI SCROLL TIMEMODE "
		"PROGLIST PROGSEL PROGSELN NEW PROGDEL STRNEW STRDEL PRINTSTR STRCPYIMM PROGGET PROGLIN1 PROGPUT PROGPUTIMM PROGNEW PROGRUN "
		"ARYNEW ARYDEL ARYSIZS ARYSET ARYGET"));
    unsigned char *code = (unsigned char *) kmallocRWX(1024 * 1024);
	KSizPtr varSp(kmemPool);
	char *cmdlin = (char *) kautoreleasePool->alloc(65536), *p;
	Env env_;
	env = &env_;
	env->lx = &lx;
	env->varSp = &varSp;
	env->code = code;
	env->qc = 0;
	env->timemode = 0;
	sub_progsel(0);
	srand((unsigned int) time(0));
	gwin = bla_openWin(256, 256, "graphics");
	for (;;) {
		bla_flushAll(gwin);
		printf("Ok\n");
skip:
		fgets(cmdlin, 65536, stdin);
		clock_t t0 = clock();
		for (p = cmdlin; *p == ' ' || *p == '\t'; p++);
		if (*p == '\n') goto skip;
		kupperCaseM(strlen(p), (unsigned char *) p, lx.chrTyp); 
		if (strncmp(p, "RUN", 3) == 0 && p[3] <= ' ') {
			if (sub_run(progseln) == 0) {
				if (env->timemode != 0)
					printf("(%.3f[sec])\n", (clock() - t0) / (double) CLOCKS_PER_SEC);
				putchar('\n');
			}
		} else if ('0' <= *p && *p <= '9') {
			int l = strtol(p, &p, 10);
			KIndexS *pn = prog_getPln(progseln);
			prog_delLine(pn, l);
			if (*p == ' ')
				p++;
			if (*p != '\n')
				prog_insLine(pn, l, strchr(p, '\n') - p, p);
			goto skip;
		} else {
			env->ln = -1;
			eval0(-1);
			KIndexS *pn = prog_getPln(progseln);
			if (evalAdd(strlen(p), p, pn) == 0) {
				eval1();
				if (env->timemode != 0)
					printf("(%.3f[sec])\n", (clock() - t0) / (double) CLOCKS_PER_SEC);
				putchar('\n');
			}
		}
	}
}
